package com.io7m.glowmaps;

import java.io.IOException;

import org.lwjgl.LWJGLException;
import org.lwjgl.input.Keyboard;
import org.lwjgl.opengl.Display;
import org.lwjgl.opengl.DisplayMode;
import org.lwjgl.opengl.GL11;
import org.lwjgl.opengl.GL13;
import org.lwjgl.opengl.GL20;
import org.lwjgl.opengl.GL30;

public final class GlowMap implements Runnable
{
  static enum RenderSelection
  {
    RENDER_ALL,
    RENDER_LIGHTS,
    RENDER_SCENE,
    RENDER_GLOW,
  }

  private static int SCREEN_WIDTH        = 640;
  private static int SCREEN_HEIGHT       = 480;
  private static int BLUR_TEXTURE_WIDTH  = GlowMap.SCREEN_WIDTH;
  private static int BLUR_TEXTURE_HEIGHT = GlowMap.SCREEN_HEIGHT;
  private static int UNIT_WIDTH          = GlowMap.SCREEN_WIDTH / 40;
  private static int UNIT_HEIGHT         = GlowMap.SCREEN_HEIGHT / 40;

  public static void main(
    final String args[])
  {
    try {
      Display.setTitle("GlowMap");
      Display.setDisplayMode(new DisplayMode(
        GlowMap.SCREEN_WIDTH,
        GlowMap.SCREEN_HEIGHT));
      Display.create();
      final GlowMap r = new GlowMap();
      r.run();
    } catch (final LWJGLException e) {
      System.err.print("fatal: LGJWL error: " + e.getMessage());
    } catch (final IOException e) {
      System.err.print("fatal: i/o error: " + e.getMessage());
    } finally {
      Display.destroy();
    }
  }

  private final int       shader_uv;
  private final int       shader_diffuse;
  private final int       shader_blur_v;
  private final int       shader_blur_h;

  private final int       tile_texture;

  private final int       fb_texture_blur_h;
  private final int       fb_texture_blur_v;
  private final int       fb_blur_h;
  private final int       fb_blur_v;
  private final int       fb_texture_scene;
  private final int       fb_scene;

  private RenderSelection selection = RenderSelection.RENDER_ALL;

  GlowMap()
    throws IOException
  {
    this.tile_texture =
      Utilities.loadTexture("dist/128.png", GL11.GL_NEAREST);

    /**
     * Create framebuffer to store horizontal pass of gaussian blur.
     */

    this.fb_texture_blur_h =
      Utilities.createEmptyTexture(
        GlowMap.BLUR_TEXTURE_WIDTH,
        GlowMap.BLUR_TEXTURE_HEIGHT);
    this.fb_blur_h = Utilities.createFramebuffer(this.fb_texture_blur_h);

    /**
     * Create framebuffer to store vertical pass of gaussian blur.
     */

    this.fb_texture_blur_v =
      Utilities.createEmptyTexture(
        GlowMap.BLUR_TEXTURE_WIDTH,
        GlowMap.BLUR_TEXTURE_HEIGHT);
    this.fb_blur_v = Utilities.createFramebuffer(this.fb_texture_blur_v);

    /**
     * Create framebuffer to store the rendered scene.
     */

    this.fb_texture_scene =
      Utilities.createEmptyTexture(
        GlowMap.SCREEN_WIDTH,
        GlowMap.SCREEN_HEIGHT);
    this.fb_scene = Utilities.createFramebuffer(this.fb_texture_scene);

    /**
     * Load shaders.
     */

    this.shader_uv =
      Utilities.createShader("dist/shader_uv.v", "dist/shader_uv.f");
    this.shader_diffuse =
      Utilities.createShader("dist/shader_plain.v", "dist/shader_plain.f");
    this.shader_blur_v =
      Utilities.createShader("dist/shader_uv.v", "dist/shader_blur_v.f");
    this.shader_blur_h =
      Utilities.createShader("dist/shader_uv.v", "dist/shader_blur_h.f");
  }

  @SuppressWarnings("static-method") private void clearFramebufferToGrey()
  {
    GL11.glViewport(0, 0, GlowMap.SCREEN_WIDTH, GlowMap.SCREEN_HEIGHT);
    GL11.glClearColor(0.1f, 0.1f, 0.1f, 1.0f);
    GL11.glClear(GL11.GL_COLOR_BUFFER_BIT);
  }

  @SuppressWarnings("static-method") private
    void
    clearFramebufferToTransparent()
  {
    GL11.glViewport(0, 0, GlowMap.SCREEN_WIDTH, GlowMap.SCREEN_HEIGHT);
    GL11.glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
    GL11.glClear(GL11.GL_COLOR_BUFFER_BIT);
  }

  private void input()
  {
    while (Keyboard.next()) {
      if (Keyboard.getEventKey() == Keyboard.KEY_R) {
        if (Keyboard.getEventKeyState()) {
          switch (this.selection) {
            case RENDER_ALL:
              this.selection = RenderSelection.RENDER_LIGHTS;
              break;
            case RENDER_LIGHTS:
              this.selection = RenderSelection.RENDER_SCENE;
              break;
            case RENDER_SCENE:
              this.selection = RenderSelection.RENDER_GLOW;
              break;
            case RENDER_GLOW:
              this.selection = RenderSelection.RENDER_ALL;
              break;
          }

          System.out.println("Rendering " + this.selection);
        }
      }
    }
  }

  private void render()
  {
    this.renderSetupCamera();

    switch (this.selection) {
      case RENDER_ALL:
      {
        this.renderEverything();
        break;
      }
      case RENDER_LIGHTS:
      {
        this.renderOnlyLights();
        break;
      }
      case RENDER_SCENE:
      {
        GL30.glBindFramebuffer(GL30.GL_FRAMEBUFFER, 0);
        this.renderOnlyScene();
        break;
      }
      case RENDER_GLOW:
      {
        GL30.glBindFramebuffer(GL30.GL_FRAMEBUFFER, this.fb_scene);
        this.renderOnlyGlow();
        break;
      }
    }
  }

  /**
   * Render both the scene and the glow map to the screen.
   */

  private void renderEverything()
  {
    /**
     * Render lights to scene buffer.
     */

    GL30.glBindFramebuffer(GL30.GL_FRAMEBUFFER, this.fb_scene);
    this.clearFramebufferToTransparent();
    this.renderLightsToCurrent();
    Utilities.checkGL();

    /**
     * Render glow to the blur_v buffer.
     */

    this.renderGlowToBlurV();

    /**
     * Render scene to scene buffer.
     */

    GL30.glBindFramebuffer(GL30.GL_FRAMEBUFFER, this.fb_scene);
    this.clearFramebufferToTransparent();
    this.renderSceneToCurrent();
    Utilities.checkGL();

    /**
     * Render composed contents of blur and scene buffers to the screen.
     */

    GL30.glBindFramebuffer(GL30.GL_FRAMEBUFFER, 0);
    this.clearFramebufferToGrey();
    Utilities.checkGL();

    GL20.glUseProgram(this.shader_uv);
    {
      GL13.glActiveTexture(GL13.GL_TEXTURE0);
      GL11.glBindTexture(GL11.GL_TEXTURE_2D, this.fb_texture_scene);
      Utilities.checkGL();

      {
        final int ut = GL20.glGetUniformLocation(this.shader_uv, "texture");
        GL20.glUniform1i(ut, 0);
        Utilities.checkGL();
        Utilities.renderScreenAlignedQuad(
          GlowMap.SCREEN_WIDTH,
          GlowMap.SCREEN_HEIGHT);
      }

      GL13.glActiveTexture(GL13.GL_TEXTURE0);
      GL11.glBindTexture(GL11.GL_TEXTURE_2D, this.fb_texture_blur_v);
      Utilities.checkGL();

      {
        final int ut = GL20.glGetUniformLocation(this.shader_uv, "texture");
        GL20.glUniform1i(ut, 0);
        Utilities.checkGL();
        Utilities.renderScreenAlignedQuad(
          GlowMap.SCREEN_WIDTH,
          GlowMap.SCREEN_HEIGHT);
      }
    }
    GL20.glUseProgram(0);
    Utilities.checkGL();
  }

  /**
   * Apply a gaussian blur to the contents of the scene buffer, using the
   * blur_h buffer to store intermediate results, and saving the results to
   * the blur_v buffer.
   */

  private void renderGlowToBlurV()
  {
    /**
     * Render scene texture into blur_h buffer with blur_h shader.
     */

    GL30.glBindFramebuffer(GL30.GL_FRAMEBUFFER, this.fb_blur_h);
    this.clearFramebufferToTransparent();
    Utilities.checkGL();

    GL20.glUseProgram(this.shader_blur_h);
    {
      GL13.glActiveTexture(GL13.GL_TEXTURE0);
      GL11.glBindTexture(GL11.GL_TEXTURE_2D, this.fb_texture_scene);
      Utilities.checkGL();

      final int ut = GL20.glGetUniformLocation(this.shader_blur_h, "texture");
      GL20.glUniform1i(ut, 0);
      Utilities.checkGL();

      final int utw =
        GL20.glGetUniformLocation(this.shader_blur_h, "texture_width");
      GL20.glUniform1f(utw, GlowMap.BLUR_TEXTURE_WIDTH);
      Utilities.checkGL();

      Utilities.renderScreenAlignedQuad(
        GlowMap.SCREEN_WIDTH,
        GlowMap.SCREEN_HEIGHT);
    }
    GL20.glUseProgram(0);
    Utilities.checkGL();

    /**
     * Render blur_h texture into blur_v buffer with blur_v shader.
     */

    GL30.glBindFramebuffer(GL30.GL_FRAMEBUFFER, this.fb_blur_v);
    this.clearFramebufferToTransparent();
    Utilities.checkGL();

    GL20.glUseProgram(this.shader_blur_v);
    {
      GL13.glActiveTexture(GL13.GL_TEXTURE0);
      GL11.glBindTexture(GL11.GL_TEXTURE_2D, this.fb_texture_blur_h);
      Utilities.checkGL();

      final int ut = GL20.glGetUniformLocation(this.shader_blur_v, "texture");
      GL20.glUniform1i(ut, 0);
      Utilities.checkGL();

      final int utw =
        GL20.glGetUniformLocation(this.shader_blur_v, "texture_height");
      GL20.glUniform1f(utw, GlowMap.BLUR_TEXTURE_HEIGHT);
      Utilities.checkGL();

      Utilities.renderScreenAlignedQuad(
        GlowMap.SCREEN_WIDTH,
        GlowMap.SCREEN_HEIGHT);
    }
    GL20.glUseProgram(0);
    Utilities.checkGL();
  }

  /**
   * Render the lights to whatever framebuffer is active.
   */

  private void renderLightsToCurrent()
  {
    GL20.glUseProgram(this.shader_diffuse);
    {
      final int ud =
        GL20.glGetUniformLocation(this.shader_diffuse, "diffuse");
      Utilities.checkGL();

      GL20.glUniform3f(ud, 1.0f, 0.0f, 0.0f);
      Utilities.renderQuadCenteredAt(-3, 3, 1, 1);

      GL20.glUniform3f(ud, 0.0f, 1.0f, 0.0f);
      Utilities.renderQuadCenteredAt(3, 3, 1, 1);

      GL20.glUniform3f(ud, 0.0f, 0.0f, 1.0f);
      Utilities.renderQuadCenteredAt(3, -3, 1, 1);

      GL20.glUniform3f(ud, 1.0f, 1.0f, 0.0f);
      Utilities.renderQuadCenteredAt(-3, -3, 1, 1);

      GL20.glUniform3f(ud, 1.0f, 1.0f, 1.0f);
      Utilities.renderQuadCenteredAt(-8, -8, 0.5, 0.5);

      GL20.glUniform3f(ud, 1.0f, 1.0f, 1.0f);
      Utilities.renderQuadCenteredAt(8, -8, 0.5, 0.5);

      GL20.glUniform3f(ud, 1.0f, 1.0f, 1.0f);
      Utilities.renderQuadCenteredAt(8, 8, 0.5, 0.5);

      GL20.glUniform3f(ud, 1.0f, 1.0f, 1.0f);
      Utilities.renderQuadCenteredAt(-8, 8, 0.5, 0.5);
    }
    GL20.glUseProgram(0);
    Utilities.checkGL();
  }

  private void renderMetalQuad()
  {
    /**
     * Render metal quad.
     */

    GL20.glUseProgram(this.shader_uv);
    {
      GL13.glActiveTexture(GL13.GL_TEXTURE0);
      GL11.glBindTexture(GL11.GL_TEXTURE_2D, this.tile_texture);
      Utilities.checkGL();

      final int ut = GL20.glGetUniformLocation(this.shader_uv, "texture");
      GL20.glUniform1i(ut, 0);
      Utilities.checkGL();

      Utilities.renderQuadCenteredAt(0, 0, 8, 8);
    }
    GL20.glUseProgram(0);
    Utilities.checkGL();
  }

  /**
   * Render only the blurred lights to the screen.
   */

  private void renderOnlyGlow()
  {
    /**
     * Render lights to scene buffer.
     */

    GL30.glBindFramebuffer(GL30.GL_FRAMEBUFFER, this.fb_scene);
    this.clearFramebufferToTransparent();
    this.renderLightsToCurrent();
    Utilities.checkGL();

    /**
     * Render glow.
     */

    this.renderGlowToBlurV();

    /**
     * Render contents of blur_v buffer to screen.
     */

    GL30.glBindFramebuffer(GL30.GL_FRAMEBUFFER, 0);
    this.clearFramebufferToGrey();
    Utilities.checkGL();

    GL20.glUseProgram(this.shader_uv);
    {
      GL13.glActiveTexture(GL13.GL_TEXTURE0);
      GL11.glBindTexture(GL11.GL_TEXTURE_2D, this.fb_texture_blur_v);
      Utilities.checkGL();

      final int ut = GL20.glGetUniformLocation(this.shader_uv, "texture");
      GL20.glUniform1i(ut, 0);
      Utilities.checkGL();
      Utilities.renderScreenAlignedQuad(
        GlowMap.SCREEN_WIDTH,
        GlowMap.SCREEN_HEIGHT);
    }
    GL20.glUseProgram(0);
    Utilities.checkGL();
  }

  /**
   * Render only the lights to the screen.
   */

  private void renderOnlyLights()
  {
    GL30.glBindFramebuffer(GL30.GL_FRAMEBUFFER, 0);
    this.clearFramebufferToGrey();
    this.renderLightsToCurrent();
  }

  /**
   * Render only the scene to the screen.
   */

  private void renderOnlyScene()
  {
    GL30.glBindFramebuffer(GL30.GL_FRAMEBUFFER, 0);
    this.clearFramebufferToGrey();
    this.renderSceneToCurrent();
  }

  /**
   * Render the scene and lights to whatever framebuffer is active.
   */

  private void renderSceneToCurrent()
  {
    this.renderMetalQuad();
    this.renderLightsToCurrent();
  }

  /**
   * Set up the standard projection and modelview matrices, and enable
   * blending.
   */

  @SuppressWarnings("static-method") private void renderSetupCamera()
  {
    GL11.glMatrixMode(GL11.GL_PROJECTION);
    GL11.glLoadIdentity();
    GL11.glFrustum(
      -GlowMap.UNIT_WIDTH,
      GlowMap.UNIT_WIDTH,
      -GlowMap.UNIT_HEIGHT,
      GlowMap.UNIT_HEIGHT,
      1.0,
      100.0);

    GL11.glMatrixMode(GL11.GL_MODELVIEW);
    GL11.glLoadIdentity();
    GL11.glTranslated(0, 0, -1);

    GL11.glEnable(GL11.GL_BLEND);
    GL11.glBlendFunc(GL11.GL_ONE, GL11.GL_ONE_MINUS_SRC_ALPHA);
  }

  public void run()
  {
    try {
      while (Display.isCloseRequested() == false) {
        this.input();
        this.render();
        Display.update();
        Display.sync(60);
      }
    } finally {
      Display.destroy();
    }
  }
}
